=begin
___________________________________________________________________________
TIG (c) 2011
All Rights Reserved.
THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
WARRANTIES,INCLUDING,WITHOUT LIMITATION,THE IMPLIED WARRANTIES OF 
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
___________________________________________________________________________
Script: Quadrilateralizer.rb
___________________________________________________________________________
Usage:
Select faces.  Note that other entity-types in the selection are ignored.
Run 'Quadrilateralizer' from the 'Plugins' menu.
If there are no faces in the selection there is an error-message, it exits.
If all selected faces are already quads then you are asked if you want to 
split them further: 'Yes' to redivide selected quad faces or No' to abort.
If the selected faces are not equivalent to 'all connected faces' then you 
are warned: 'Yes' to continue, 'No' to abort.  It is recommended that all 
connected faces are processed together [treble-click a face to select all] 
to ensure that all connected faces become quadrilaterals; otherwise later 
processing of these non-quad faces could result in the connected quads 
being made into non-quads, or if all connected faces were selected the 
non-quads would split but the previously made quads will be split again - 
it could get messy - this might be what is required... but usually 
processing all connected at once is the best method.
All selected faces are now split into quadrilaterals ['quads'].
Note that any preexisting quads in the selection will also be split 
because splitting any adjacent non-quads will force these existing quads 
into non-quads anyway.
UV texture mapping etc is kept.
It is one step undo-able.
The vertices of these quad faces can be adjusted with the Move tool...
___________________________________________________________________________
Donations: By PayPal.com to info @ revitrev.org
___________________________________________________________________________
Version:
1.0 20110214 First Release.
1.1 20110215 More info and percentage progress reported in Status Bar.
___________________________________________________________________________
=end
###
require 'sketchup.rb'
###
class Quadrilateralizer

def initialize()
  Sketchup::set_status_text("Quadrilateralizer: Analyzing.",SB_PROMPT)
  @model=Sketchup.active_model
  @ents=@model.active_entities
  ss=@model.selection
  faces=[]
  ss.each{|e|faces << e if e.kind_of?(Sketchup::Face)}
  if not faces[0]
    UI.messagebox("Quadrilateralizer:\n\nThere are NO Faces in the Selection !")
    return nil
  end#if
  ### check if needs processing
  allquads=true
  faces.each{|face|
    if face.edges.length!=4
      allquads=false
      break
    end#if
  }
  if allquads ###MB_YESNO - Contains YES (6) and NO (7)
    UI.beep
    return nil if UI.messagebox("Quadrilateralizer:\n\nALL Selected Faces are Quads.\nDo you want to Split them further ?",MB_YESNO)==7
  end#if
  Sketchup::set_status_text("Quadrilateralizer: Analyzing..",SB_PROMPT)
  connected=[]
  faces.each{|face|
    conn=[]
    face.all_connected.each{|e| conn << e.to_s if e.kind_of?(Sketchup::Face)}
    connected << conn
  }
  connected.flatten!
  connected.uniq!
  connected.sort!
  ffaces=[]
  faces.each{|face|ffaces << face.to_s}
  ffaces.uniq!
  ffaces.sort!
  Sketchup::set_status_text("Quadrilateralizer: Analyzing...",SB_PROMPT)
  if ffaces != connected ###MB_YESNO - Contains YES (6) and NO (7)
    UI.beep
    return nil if UI.messagebox("Quadrilateralizer:\n\nThe Selected Faces are NOT\nALL of the Connected Faces.\nDo you want to Continue ?",MB_YESNO)==7
  end#if
  ###
  ss.clear
 begin
    @model.start_operation("Quadrilateralizer",true)
 rescue
    @model.start_operation("Quadrilateralizer")
 end
  Sketchup::set_status_text("Quadrilateralizer: Analyzing.....",SB_PROMPT)
  @gp=@ents.add_group()
  @tents=@gp.entities
  msg1="Quadrilateralizing "
  num=0
  tot=faces.length.to_f
  faces.each{|face|
    next if not face.valid?
    per=(100.0*num.to_f/tot).to_i.to_s
    msg=msg1+per+"%"
    Sketchup::set_status_text(msg,SB_PROMPT)
    if face.edges.length==3
      self.quadrilateralate(face)
    else
      faces=self.triangulate(face)
      faces.each{|fac|self.quadrilateralate(fac)if fac.valid?}
    end#if
    num+=1
  }
  if @ents.parent==@model
    tr=Geom::Transformation.new
  else
    tr=@ents.parent.instances[0].transformation
  end#if
  Sketchup::set_status_text("Quadrilateralizer: Splitting...",SB_PROMPT)
  @ents.intersect_with(true, tr, @ents, tr, true, @tents.to_a)
  @gp.erase!
  Sketchup::set_status_text("",SB_PROMPT)
 @model.commit_operation
  return nil
  ###
end#def

def quadrilateralate(face)
  edges=face.edges
  e0=edges[0]
  e1=edges[1]
  e2=edges[2]
  p0=e0.line[0]
  p1=e1.line[0]
  p2=e2.line[0]
  v0=e0.line[1]
  v1=e1.line[1]
  v2=e2.line[1]
  d0=e0.length
  d1=e1.length
  d2=e2.length
  c0=p0.offset(v0,d0/2)
  c1=p1.offset(v1,d1/2)
  c2=p2.offset(v2,d2/2)
  nn=face.normal
  ps=[]
  face.vertices.each{|v| ps << v.position}
  p0=ps[0]
  p1=ps[1]
  p2=ps[2]
  v01=p0.vector_to(p1)
  v02=p0.vector_to(p2)
  a01=v01.angle_between(v02)
  t01=Geom::Transformation.rotation(p0,nn,a01/2.0)
  v01.transform!(t01)
  v10=p1.vector_to(p0)
  v12=p1.vector_to(p2)
  a10=v10.angle_between(v12)
  t10=Geom::Transformation.rotation(p1,nn,a10/-2.0)
  v10.transform!(t10)
  l01=[p0,v01]
  l10=[p1,v10]
  cp=Geom.intersect_line_line(l01,l10)
  @tents.add_line(c0,cp)
  @tents.add_line(c1,cp)
  @tents.add_line(c2,cp)
  ###
end#def

def triangulate(face)
 ### if already a triangle leave alone...
 if face.vertices[3]
  mesh=face.mesh(3)
  ents=face.parent.entities
  mat=face.material
  bac=face.back_material
   # (TT) # Instead of sampling vertices we sample points of the face's plane.
   # This is because we need four points, and some times a face only has three vertices.
   # And we also get the wrong result if any of the first four vertices are co-linear.
   samples = []
   samples << face.vertices[0].position             # 0,0 | Origin
   samples << samples[0].offset(face.normal.axes.x) # 1,0 | Offset Origin in X
   samples << samples[0].offset(face.normal.axes.y) # 0,1 | Offset Origin in Y
   samples << samples[1].offset(face.normal.axes.y) # 1,1 | Offset X in Y
   ### get uvs
   tw=Sketchup.create_texture_writer
   # (TT)
   uv_helper=face.get_UVHelper(true,true,tw) # (TT) Only need one UV Helper
   xyz = []
   uv_f = []
   uv_b = []
   samples.each { |position|
      # XYZ 3D coordinates
      xyz << position
      # UV 2D coordinates
      # Front
      uvq = uv_helper.get_front_UVQ(position)
      uv_f << self.flattenUVQ(uvq)
      # Back
      uvq = uv_helper.get_back_UVQ(position)
      uv_b << self.flattenUVQ(uvq)
   }
  ###
  grp=ents.add_group
  grp.entities.fill_from_mesh(mesh, true, 0)
  # (TT) Adds the mesh without soft/smooth edges. (and faster)
  grp.entities.to_a.each{|e|
    next unless e.class==Sketchup::Face
    if mat # (TT) # Position texture.
      pts = []
      (0..3).each { |i|
         pts << xyz[i]
         pts << uv_f[i]
      }
        e.position_material(mat,pts, true)
    end#if
    if bac # (TT) # Position texture.
      pts = []
      (0..3).each { |i|
         pts << xyz[i]
         pts << uv_b[i]
      }
       e.position_material(bac, pts, false)
    end#if
  }
  ### force intersection of all triangles so if any 4 sided they are fixed.
  grp.entities.intersect_with(true, grp.transformation, grp.entities, grp.transformation, true, grp.entities.to_a)
  ### fix 4 sided triangles...
  faces=[]
  grp.entities.to_a.each{|e|
    next unless e.class==Sketchup::Face
    faces << e
    verts=e.vertices
    if verts.length > 3
      verts.each{|v|
           (verts-[v]).each{|vv|
                grp.entities.add_line(v.position, vv.position)
            }
        }
    end#if
  }
  face.erase! if face.valid?
  grp.explode if grp.valid?
  return faces
 end#if
 
end#def

# Get UV coordinates from UVQ matrix. (TT)
def flattenUVQ(uvq)
  return Geom::Point3d.new(uvq.x / uvq.z, uvq.y / uvq.z, 1.0)
end
###

end#classs Quadizer
###
### Menu ###------------------------------------------------------------
if not file_loaded?(File.basename(__FILE__))
  UI.menu("Plugins").add_item("Quadrilateralizer"){Quadrilateralizer.new()}
end#if
file_loaded(File.basename(__FILE__))
###
